#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_conditional_advertisement.py
#
# Copyright (c) 2020 by
# Samsung R&D Institute India - Bangalore.
# Madhurilatha Kuruganti
#

"""
Test BGP conditional advertisement functionality.

    +--------+            +--------+            +--------+
    |        |            |        |            |        |
    |   R1   |------------|   R2   |------------|   R3   |
    |        |            |        |            |        |
    +--------+            +--------+            +--------+

R2 is DUT and peers with R1 and R3 in default bgp instance.

Following tests are covered under BGP conditional advertisement functionality.
Conditional advertisement
-------------------------
TC11: R3 BGP convergence, without advertise-map configuration.
      All routes are advertised to R3.

TC21: exist-map routes present in R2's BGP table.
      advertise-map routes present in R2's BGP table are advertised to R3.
TC22: exist-map routes not present in R2's BGP table
      advertise-map routes present in R2's BGP table are withdrawn from R3.
TC23: advertise-map with exist-map configuration is removed from a peer
      send normal BGP update to advertise previously withdrawn routes if any.

TC31: non-exist-map routes not present in R2's BGP table
      advertise-map routes present in R2's BGP table are advertised to R3.
TC32: non-exist-map routes present in R2's BGP table
      advertise-map routes present in R2's BGP table are withdrawn from R3.
TC33: advertise-map with non-exist-map configuration is removed from a peer
      send normal BGP update to advertisepreviously withdrawn routes if any.

TC41: non-exist-map route-map configuration removed in R2.
      advertise-map routes present in R2's BGP table are advertised to R3.
TC42: exist-map route-map configuration removed in R2
      advertise-map routes present in R2's BGP table are withdrawn from R3.

Conditional advertisement(received routes) along with Route-map Filter
----------------------------------------------------------------------
TC51: exist-map routes present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 except advertise-map routes.
TC52: exist-map routes present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 including advertise-map routes.
TC53: non-exist-map routes present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 including advertise-map routes.
TC54: non-exist-map routes present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 except advertise-map routes.

TC61: exist-map routes not present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 including advertise-map routes.
TC62: exist-map routes not present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 except advertise-map routes.
TC63: non-exist-map routes not present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 except advertise-map routes.
TC64: non-exist-map routes not present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 including advertise-map routes.

Conditional advertisement(attached routes) along with Route-map Filter
-----------------------------------------------------------------
TC71: exist-map routes present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 except advertise-map routes.
TC72: exist-map routes present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 including advertise-map routes.
TC73: non-exist-map routes present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 including advertise-map routes.
TC74: non-exist-map routes present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 except advertise-map routes.

TC81: exist-map routes not present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 including advertise-map routes.
TC82: exist-map routes not present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 except advertise-map routes.
TC83: non-exist-map routes not present in R2's BGP table, with route-map filter.
      All routes are withdrawn from R3 except advertise-map routes.
TC84: non-exist-map routes not present in R2's BGP table, without route-map filter.
      All routes are advertised to R3 including advertise-map routes.

TC91: exist-map routes present in R2's BGP table, with route-map filter and network.
      All routes are advertised to R3 including advertise-map routes.
TC92: exist-map routes present in R2's BGP table, with route-map filter and no network.
      All routes are advertised to R3 except advertise-map routes.
TC93: non-exist-map routes not present in R2's BGP table, with route-map filter and network.
      All routes are advertised to R3 including advertise-map routes.
TC94: non-exist-map routes not present in R2's BGP table, with route-map filter and no network.
      All routes are advertised to R3 except advertise-map routes.

i.e.
+----------------+-------------------------+------------------------+
|  Routes in     |  exist-map status       | advertise-map status   |
|  BGP table     |                         |                        |
+----------------+-------------------------+------------------------+
|  Present       |  Condition matched      | Advertise              |
+----------------+-------------------------+------------------------+
|  Not Present   |  Condition not matched  | Withdrawn              |
+----------------+-------------------------+------------------------+
|                |  non-exist-map status   | advertise-map status   |
|                |                         |                        |
+----------------+-------------------------+------------------------+
|  Present       |  Condition matched      | Withdrawn              |
+----------------+-------------------------+------------------------+
|  Not Present   |  Condition not matched  | Advertise              |
+----------------+-------------------------+------------------------+
Here in this topology, based on the default route presence in R2 and
the configured condition-map (exist-map/non-exist-map) 10.139.224.0/20
will be either advertised/withdrawn to/from R3.
"""

import os
import sys
import json
import time
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    r1 = tgen.add_router("r1")
    r2 = tgen.add_router("r2")
    r3 = tgen.add_router("r3")

    switch = tgen.add_switch("s1")
    switch.add_link(r1)
    switch.add_link(r2)

    switch = tgen.add_switch("s2")
    switch.add_link(r2)
    switch.add_link(r3)


def setup_module(mod):
    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()

    logger.info("Running setup_module() done")


def teardown_module(mod):
    """
    Teardown the pytest environment
    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


def all_routes_advertised(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": [{"protocol": "bgp"}],
        "192.0.2.1/32": [{"protocol": "bgp"}],
        "192.0.2.5/32": [{"protocol": "bgp"}],
        "10.139.224.0/20": [{"protocol": "bgp"}],
        "203.0.113.1/32": [{"protocol": "bgp"}],
    }
    return topotest.json_cmp(output, expected)


def all_routes_withdrawn(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": None,
        "192.0.2.1/32": None,
        "192.0.2.5/32": None,
        "10.139.224.0/20": None,
        "203.0.113.1/32": None,
    }
    return topotest.json_cmp(output, expected)


def default_route_withdrawn(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": None,
        "192.0.2.1/32": [{"protocol": "bgp"}],
        "192.0.2.5/32": [{"protocol": "bgp"}],
        "10.139.224.0/20": [{"protocol": "bgp"}],
        "203.0.113.1/32": [{"protocol": "bgp"}],
    }
    return topotest.json_cmp(output, expected)


# BGP conditional advertisement with route-maps
# EXIST-MAP, ADV-MAP-1 and RMAP-1
def exist_map_routes_present(router):
    return all_routes_advertised(router)


def exist_map_routes_not_present(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": None,
        "192.0.2.1/32": None,
        "192.0.2.5/32": [{"protocol": "bgp"}],
        "10.139.224.0/20": None,
        "203.0.113.1/32": [{"protocol": "bgp"}],
    }
    return topotest.json_cmp(output, expected)


def non_exist_map_routes_present(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": [{"protocol": "bgp"}],
        "192.0.2.1/32": None,
        "192.0.2.5/32": [{"protocol": "bgp"}],
        "10.139.224.0/20": None,
        "203.0.113.1/32": [{"protocol": "bgp"}],
    }
    return topotest.json_cmp(output, expected)


def non_exist_map_routes_not_present(router):
    return default_route_withdrawn(router)


def exist_map_no_condition_route_map(router):
    return non_exist_map_routes_present(router)


def non_exist_map_no_condition_route_map(router):
    return all_routes_advertised(router)


def exist_map_routes_present_rmap_filter(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": None,
        "192.0.2.1/32": [{"protocol": "bgp"}],
        "192.0.2.5/32": None,
        "10.139.224.0/20": [{"protocol": "bgp"}],
        "203.0.113.1/32": None,
    }
    return topotest.json_cmp(output, expected)


def exist_map_routes_present_no_rmap_filter(router):
    return all_routes_advertised(router)


def non_exist_map_routes_present_rmap_filter(router):
    return all_routes_withdrawn(router)


def non_exist_map_routes_present_no_rmap_filter(router):
    return non_exist_map_routes_present(router)


def exist_map_routes_not_present_rmap_filter(router):
    return all_routes_withdrawn(router)


def exist_map_routes_not_present_no_rmap_filter(router):
    return exist_map_routes_not_present(router)


def non_exist_map_routes_not_present_rmap_filter(router):
    return exist_map_routes_present_rmap_filter(router)


def non_exist_map_routes_not_present_no_rmap_filter(router):
    return non_exist_map_routes_not_present(router)


# BGP conditional advertisement with route-maps
# EXIST-MAP, ADV-MAP-2 and RMAP-2
def exist_map_routes_not_present_rmap2_filter(router):
    return all_routes_withdrawn(router)


def exist_map_routes_not_present_no_rmap2_filter(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": None,
        "192.0.2.1/32": [{"protocol": "bgp"}],
        "192.0.2.5/32": [{"protocol": "bgp"}],
        "10.139.224.0/20": [{"protocol": "bgp"}],
        "203.0.113.1/32": None,
    }
    return topotest.json_cmp(output, expected)


def non_exist_map_routes_not_present_rmap2_filter(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": None,
        "192.0.2.1/32": None,
        "192.0.2.5/32": None,
        "10.139.224.0/20": None,
        "203.0.113.1/32": [{"protocol": "bgp", "metric": 911}],
    }
    return topotest.json_cmp(output, expected)


def non_exist_map_routes_not_present_no_rmap2_filter(router):
    return non_exist_map_routes_not_present(router)


def exist_map_routes_present_rmap2_filter(router):
    return non_exist_map_routes_not_present_rmap2_filter(router)


def exist_map_routes_present_no_rmap2_filter(router):
    return all_routes_advertised(router)


def non_exist_map_routes_present_rmap2_filter(router):
    return all_routes_withdrawn(router)


def non_exist_map_routes_present_no_rmap2_filter(router):
    output = json.loads(router.vtysh_cmd("show ip route json"))
    expected = {
        "0.0.0.0/0": [{"protocol": "bgp"}],
        "192.0.2.1/32": [{"protocol": "bgp"}],
        "192.0.2.5/32": [{"protocol": "bgp"}],
        "10.139.224.0/20": [{"protocol": "bgp"}],
        "203.0.113.1/32": None,
    }
    return topotest.json_cmp(output, expected)


def exist_map_routes_present_rmap2_network(router):
    return non_exist_map_routes_not_present_rmap2_filter(router)


def exist_map_routes_present_rmap2_no_network(router):
    return all_routes_withdrawn(router)


def non_exist_map_routes_not_present_rmap2_network(router):
    return non_exist_map_routes_not_present_rmap2_filter(router)


def non_exist_map_routes_not_present_rmap2_no_network(router):
    return all_routes_withdrawn(router)


passed = "PASSED!!!"
failed = "FAILED!!!"


def test_bgp_conditional_advertisement_tc_1_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC11: R3 BGP convergence, without advertise-map configuration.
    # All routes are advertised to R3.
    test_func = functools.partial(all_routes_advertised, router3)
    success, result = topotest.run_and_expect(test_func, None, count=130, wait=1)

    msg = 'TC11: "router3" BGP convergence - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_2_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC21: exist-map routes present in R2's BGP table.
    # advertise-map routes present in R2's BGP table are advertised to R3.
    router2.vtysh_cmd(
        """
          configure terminal
            router bgp 2
              address-family ipv4 unicast
               neighbor 10.10.20.3 advertise-map ADV-MAP-1 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(exist_map_routes_present, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = 'TC21: exist-map routes present in "router2" BGP table - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_2_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC22: exist-map routes not present in R2's BGP table
    # advertise-map routes present in R2's BGP table are withdrawn from R3.
    router1.vtysh_cmd(
        """
          configure terminal
            router bgp 1
              address-family ipv4 unicast
               no network 0.0.0.0/0 route-map DEF
        """
    )

    test_func = functools.partial(exist_map_routes_not_present, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = 'TC22: exist-map routes not present in "router2" BGP table - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_2_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC23: advertise-map with exist-map configuration is removed from a peer
    # send normal BGP update to advertise previously withdrawn routes if any.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 advertise-map ADV-MAP-1 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(default_route_withdrawn, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC23: advertise-map with exist-map configuration is removed from peer - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_3_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC31: non-exist-map routes not present in R2's BGP table
    # advertise-map routes present in R2's BGP table are advertised to R3.
    router2.vtysh_cmd(
        """
          configure terminal
            router bgp 2
              address-family ipv4 unicast
               neighbor 10.10.20.3 advertise-map ADV-MAP-1 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(non_exist_map_routes_not_present, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = 'TC31: non-exist-map routes not present in "router2" BGP table - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_3_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC32: non-exist-map routes present in R2's BGP table
    # advertise-map routes present in R2's BGP table are withdrawn from R3.
    router1.vtysh_cmd(
        """
          configure terminal
            router bgp 1
              address-family ipv4 unicast
               network 0.0.0.0/0 route-map DEF
        """
    )

    test_func = functools.partial(non_exist_map_routes_present, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = 'TC32: non-exist-map routes present in "router2" BGP table - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_3_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC33: advertise-map with non-exist-map configuration is removed from a peer
    # send normal BGP update to advertisepreviously withdrawn routes if any.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 advertise-map ADV-MAP-1 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(all_routes_advertised, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = (
        "TC33: advertise-map with non-exist-map configuration is removed from a peer - "
    )
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_4_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC41: non-exist-map route-map configuration removed in R2.
    # advertise-map routes present in R2's BGP table are advertised to R3.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 advertise-map ADV-MAP-1 non-exist-map EXIST-MAP
           no route-map EXIST-MAP permit 10
        """
    )

    test_func = functools.partial(non_exist_map_no_condition_route_map, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = 'TC41: non-exist-map route-map removed in "router2" - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_4_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC42: exist-map route-map configuration removed in R2
    # advertise-map routes present in R2's BGP table are withdrawn from R3.
    router2.vtysh_cmd(
        """
          configure terminal
            router bgp 2
              address-family ipv4 unicast
               neighbor 10.10.20.3 advertise-map ADV-MAP-1 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(exist_map_no_condition_route_map, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = 'TC42: exist-map route-map removed in "router2" - '
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_5_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC51: exist-map routes present in R2's BGP table, with route-map filter.
    # All routes are withdrawn from R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           route-map EXIST-MAP permit 10
            match community DEFAULT-ROUTE
            match ip address prefix-list DEFAULT-ROUTE
           !
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-1 out
        """
    )

    test_func = functools.partial(exist_map_routes_present_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC51: exist-map routes present with route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_5_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC52: exist-map routes present in R2's BGP table, no route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-1 out
        """
    )

    test_func = functools.partial(exist_map_routes_present_no_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC52: exist-map routes present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_5_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC53: non-exist-map routes present in R2's BGP table, with route-map filter.
    # All routes are withdrawn from R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
            router bgp 2
              address-family ipv4 unicast
               neighbor 10.10.20.3 route-map RMAP-1 out
               neighbor 10.10.20.3 advertise-map ADV-MAP-1 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(non_exist_map_routes_present_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC53: non-exist-map routes present, with route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_5_4():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC54: non-exist-map routes present in R2's BGP table, no route-map filter.
    # All routes are advertised to R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
            router bgp 2
              address-family ipv4 unicast
               no neighbor 10.10.20.3 route-map RMAP-1 out
        """
    )

    test_func = functools.partial(non_exist_map_routes_present_no_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC54: non-exist-map routes present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_6_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC61: exist-map routes not present in R2's BGP table, with route-map filter.
    # All routes are withdrawn from R3 including advertise-map routes.
    router1.vtysh_cmd(
        """
          configure terminal
            router bgp 1
              address-family ipv4 unicast
               no network 0.0.0.0/0 route-map DEF
        """
    )
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-1 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-1 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(exist_map_routes_not_present_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC61: exist-map routes not present, route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_6_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC62: exist-map routes not present in R2's BGP table, without route-map filter.
    # All routes are advertised to R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-1 out
        """
    )

    test_func = functools.partial(exist_map_routes_not_present_no_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC62: exist-map routes not present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_6_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC63: non-exist-map routes not present in R2's BGP table, with route-map filter.
    # All routes are withdrawn from R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-1 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-1 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(non_exist_map_routes_not_present_rmap_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC63: non-exist-map routes not present, route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_6_4():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC64: non-exist-map routes not present in R2's BGP table, without route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-1 out
        """
    )

    test_func = functools.partial(
        non_exist_map_routes_not_present_no_rmap_filter, router3
    )
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC64: non-exist-map routes not present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_7_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC71: exist-map routes present in R2's BGP table, with route-map filter.
    # All routes are withdrawn from R3 except advertise-map routes.
    router1.vtysh_cmd(
        """
          configure terminal
           router bgp 1
            address-family ipv4 unicast
             network 0.0.0.0/0 route-map DEF
        """
    )
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-2 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-2 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(exist_map_routes_present_rmap2_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC71: exist-map routes present, route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_7_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC72: exist-map routes present in R2's BGP table, without route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-2 out
        """
    )

    test_func = functools.partial(exist_map_routes_present_no_rmap2_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC72: exist-map routes present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_7_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC73: non-exist-map routes present in R2's BGP table, with route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-2 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-2 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(non_exist_map_routes_present_rmap2_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC73: non-exist-map routes present, route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_7_4():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC74: non-exist-map routes present in R2's BGP table, without route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-2 out
        """
    )

    test_func = functools.partial(non_exist_map_routes_present_no_rmap2_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC74: non-exist-map routes present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_8_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC81: exist-map routes not present in R2's BGP table, with route-map filter.
    # All routes are withdrawn from R3 including advertise-map routes.
    router1.vtysh_cmd(
        """
          configure terminal
           router bgp 1
            address-family ipv4 unicast
             no network 0.0.0.0/0 route-map DEF
        """
    )
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-2 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-2 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(exist_map_routes_not_present_rmap2_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC81: exist-map routes not present, route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_8_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC82: exist-map routes not present in R2's BGP table, without route-map filter.
    # All routes are advertised to R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-2 out
        """
    )

    test_func = functools.partial(exist_map_routes_not_present_no_rmap2_filter, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC82: exist-map routes not present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_8_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC83: non-exist-map routes not present in R2's BGP table, with route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-2 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-2 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(
        non_exist_map_routes_not_present_rmap2_filter, router3
    )
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC83: non-exist-map routes not present, route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_8_4():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC84: non-exist-map routes not present in R2's BGP table, without route-map filter.
    # All routes are advertised to R3 including advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no neighbor 10.10.20.3 route-map RMAP-2 out
        """
    )

    test_func = functools.partial(
        non_exist_map_routes_not_present_no_rmap2_filter, router3
    )
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC84: non-exist-map routes not present, no route-map filter - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_9_1():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC91: exist-map routes present in R2's BGP table, with route-map filter and network.
    # All routes are advertised to R3 including advertise-map routes.
    router1.vtysh_cmd(
        """
          configure terminal
           router bgp 1
            address-family ipv4 unicast
             network 0.0.0.0/0 route-map DEF
        """
    )
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             neighbor 10.10.20.3 route-map RMAP-2 out
             neighbor 10.10.20.3 advertise-map ADV-MAP-2 exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(exist_map_routes_present_rmap2_network, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC91: exist-map routes present, route-map filter and network - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_9_2():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC92: exist-map routes present in R2's BGP table, with route-map filter and no network.
    # All routes are advertised to R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no network 203.0.113.1/32
        """
    )

    test_func = functools.partial(exist_map_routes_present_rmap2_no_network, router3)
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC92: exist-map routes present, route-map filter and no network - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_9_3():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC93: non-exist-map routes not present in R2's BGP table, with route-map filter and network.
    # All routes are advertised to R3 including advertise-map routes.
    router1.vtysh_cmd(
        """
          configure terminal
           router bgp 1
            address-family ipv4 unicast
             no network 0.0.0.0/0 route-map DEF
        """
    )
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             network 203.0.113.1/32
             neighbor 10.10.20.3 advertise-map ADV-MAP-2 non-exist-map EXIST-MAP
        """
    )

    test_func = functools.partial(
        non_exist_map_routes_not_present_rmap2_network, router3
    )
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC93: non-exist-map routes not present, route-map filter and network - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_bgp_conditional_advertisement_tc_9_4():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router1 = tgen.gears["r1"]
    router2 = tgen.gears["r2"]
    router3 = tgen.gears["r3"]

    # TC94: non-exist-map routes not present in R2's BGP table, with route-map filter and no network.
    # All routes are advertised to R3 except advertise-map routes.
    router2.vtysh_cmd(
        """
          configure terminal
           router bgp 2
            address-family ipv4 unicast
             no network 203.0.113.1/32
        """
    )

    test_func = functools.partial(
        non_exist_map_routes_not_present_rmap2_no_network, router3
    )
    _, result = topotest.run_and_expect(test_func, None, count=90, wait=1)

    msg = "TC94: non-exist-map routes not present, route-map filter and no network - "
    assert result is None, msg + failed

    logger.info(msg + passed)


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2021 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test BGP VRF Lite:
1. Verify that locally imported routes are selected as best path over eBGP imported routes
    peers.
2. Verify ECMP for imported routes from different VRFs.
"""

import os
import sys
import time
import pytest
import platform

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# Required to instantiate the topology builder class.

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen
from lib.topotest import version_cmp

from lib.common_config import (
    start_topology,
    write_test_header,
    check_address_types,
    write_test_footer,
    reset_config_on_routers,
    verify_rib,
    step,
    create_static_routes,
    check_router_status,
    apply_raw_config,
)

from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_bgp_rib,
    verify_bgp_bestpath,
)
from lib.topojson import build_config_from_json

pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]

# Global variables
NETWORK1_1 = {"ipv4": "11.11.11.1/32", "ipv6": "11:11::1/128"}
NETWORK1_2 = {"ipv4": "11.11.11.11/32", "ipv6": "11:11::11/128"}
NETWORK1_3 = {"ipv4": "10.10.10.1/32", "ipv6": "10:10::1/128"}
NETWORK1_4 = {"ipv4": "10.10.10.100/32", "ipv6": "10:10::100/128"}
NETWORK1_5 = {"ipv4": "110.110.110.1/32", "ipv6": "110:110::1/128"}
NETWORK1_6 = {"ipv4": "110.110.110.100/32", "ipv6": "110:110::100/128"}

NETWORK2_1 = {"ipv4": "22.22.22.2/32", "ipv6": "22:22::2/128"}
NETWORK2_2 = {"ipv4": "22.22.22.22/32", "ipv6": "22:22::22/128"}
NETWORK2_3 = {"ipv4": "20.20.20.20/32", "ipv6": "20:20::20/128"}
NETWORK2_4 = {"ipv4": "20.20.20.200/32", "ipv6": "20:20::200/128"}
NETWORK2_5 = {"ipv4": "220.220.220.20/32", "ipv6": "220:220::20/128"}
NETWORK2_6 = {"ipv4": "220.220.220.200/32", "ipv6": "220:220::200/128"}

NETWORK3_1 = {"ipv4": "30.30.30.3/32", "ipv6": "30:30::3/128"}
NETWORK3_2 = {"ipv4": "30.30.30.30/32", "ipv6": "30:30::30/128"}

PREFIX_LIST = {
    "ipv4": ["11.11.11.1", "22.22.22.2", "22.22.22.22"],
    "ipv6": ["11:11::1", "22:22::2", "22:22::22"],
}
PREFERRED_NEXT_HOP = "global"
VRF_LIST = ["RED", "BLUE", "GREEN"]
COMM_VAL_1 = "100:100"
COMM_VAL_2 = "500:500"
COMM_VAL_3 = "600:600"
BESTPATH = {"ipv4": "0.0.0.0", "ipv6": "::"}


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_vrf_lite_best_path_topo2.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Run these tests for kernel version 4.19 or above
    if version_cmp(platform.release(), "4.19") < 0:
        error_msg = (
            "BGP vrf dynamic route leak tests will not run "
            '(have kernel "{}", but it requires >= 4.19)'.format(platform.release())
        )
        pytest.skip(error_msg)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    global BGP_CONVERGENCE
    global ADDR_TYPES
    ADDR_TYPES = check_address_types()

    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "setup_module : Failed \n Error: {}".format(
        BGP_CONVERGENCE
    )

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
#
#   Testcases
#
#####################################################


def test_dynamic_import_ecmp_imported_routed_diffrent_vrfs_p0(request):
    """
    Verify ECMP for imported routes from different VRFs.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)
    reset_config_on_routers(tgen)

    step(
        "Configure same static routes in tenant vrfs RED and GREEN on router "
        "R3 and redistribute in respective BGP process"
    )

    for vrf_name in ["RED", "GREEN"]:
        for addr_type in ADDR_TYPES:
            if vrf_name == "GREEN":
                next_hop_vrf = topo["routers"]["r1"]["links"]["r3-link3"][
                    addr_type
                ].split("/")[0]
            else:
                next_hop_vrf = topo["routers"]["r2"]["links"]["r3-link1"][
                    addr_type
                ].split("/")[0]
            static_routes = {
                "r3": {
                    "static_routes": [
                        {
                            "network": [NETWORK1_1[addr_type]],
                            "next_hop": next_hop_vrf,
                            "vrf": vrf_name,
                        }
                    ]
                }
            }

            result = create_static_routes(tgen, static_routes)
            assert result is True, "Testcase {} :Failed \n Error: {}".format(
                tc_name, result
            )

        step("Redistribute static route on BGP VRF : {}".format(vrf_name))
        temp = {}
        for addr_type in ADDR_TYPES:
            temp.update(
                {addr_type: {"unicast": {"redistribute": [{"redist_type": "static"}]}}}
            )

        redist_dict = {
            "r3": {"bgp": [{"vrf": vrf_name, "local_as": 3, "address_family": temp}]}
        }

        result = create_router_bgp(tgen, topo, redist_dict)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step(
        "Verify that configured static routes are installed in respective "
        "BGP table for vrf RED & GREEN"
    )
    for vrf_name in ["RED", "GREEN"]:
        for addr_type in ADDR_TYPES:
            if vrf_name == "GREEN":
                next_hop_vrf = topo["routers"]["r1"]["links"]["r3-link3"][
                    addr_type
                ].split("/")[0]
            else:
                next_hop_vrf = topo["routers"]["r2"]["links"]["r3-link1"][
                    addr_type
                ].split("/")[0]
            static_routes = {
                "r3": {
                    "static_routes": [
                        {"network": [NETWORK1_1[addr_type]], "vrf": vrf_name}
                    ]
                }
            }

            result = verify_bgp_rib(
                tgen, addr_type, "r3", static_routes, next_hop=next_hop_vrf
            )
            assert result is True, "Testcase {} : Failed \n Error {}".format(
                tc_name, result
            )

            result = verify_rib(
                tgen, addr_type, "r3", static_routes, next_hop=next_hop_vrf
            )
            assert result is True, "Testcase {} : Failed \n Error {}".format(
                tc_name, result
            )

    step("Import vrf RED and GREEN into default vrf and Configure ECMP")
    bgp_val = []
    for vrf_name in ["RED", "GREEN"]:
        temp = {}
        for addr_type in ADDR_TYPES:
            temp.update(
                {
                    addr_type: {
                        "unicast": {
                            "import": {"vrf": vrf_name},
                            "maximum_paths": {"ebgp": 2},
                        }
                    }
                }
            )

        bgp_val.append({"local_as": 3, "address_family": temp})

    import_dict = {"r3": {"bgp": bgp_val}}

    result = create_router_bgp(tgen, topo, import_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("Configure bgp bestpath on router r3")
    r3_raw_config = {
        "r3": {"raw_config": ["router bgp 3", "bgp bestpath as-path multipath-relax"]}
    }
    result = apply_raw_config(tgen, r3_raw_config)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step(
        "Verify that routes are imported with two different next-hop vrfs "
        "and IPs. Additionally R3 must do ECMP for both the routes."
    )

    for addr_type in ADDR_TYPES:
        next_hop_vrf = [
            topo["routers"]["r2"]["links"]["r3-link1"][addr_type].split("/")[0],
            topo["routers"]["r1"]["links"]["r3-link3"][addr_type].split("/")[0],
        ]
        static_routes = {
            "r3": {
                "static_routes": [
                    {
                        "network": [NETWORK1_1[addr_type]],
                    }
                ]
            }
        }

        result = verify_bgp_rib(
            tgen, addr_type, "r3", static_routes, next_hop=next_hop_vrf
        )
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

        result = verify_rib(tgen, addr_type, "r3", static_routes, next_hop=next_hop_vrf)
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

    step(
        "Now change the next-hop of static routes in vrf RED and GREEN to "
        "same IP address"
    )
    for addr_type in ADDR_TYPES:
        next_hop_vrf = topo["routers"]["r1"]["links"]["r3-link3"][addr_type].split("/")[
            0
        ]
        static_routes = {
            "r3": {
                "static_routes": [
                    {
                        "network": [NETWORK1_1[addr_type]],
                        "next_hop": next_hop_vrf,
                        "vrf": "RED",
                    },
                    {
                        "network": [NETWORK1_1[addr_type]],
                        "next_hop": topo["routers"]["r2"]["links"]["r3-link1"][
                            addr_type
                        ].split("/")[0],
                        "vrf": "RED",
                        "delete": True,
                    },
                ]
            }
        }

        result = create_static_routes(tgen, static_routes)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    step(
        "Verify that now routes are imported with two different next-hop "
        "vrfs but same IPs. Additionally R3 must do ECMP for both the routes"
    )

    for addr_type in ADDR_TYPES:
        next_hop_vrf = [
            topo["routers"]["r1"]["links"]["r3-link3"][addr_type].split("/")[0],
            topo["routers"]["r1"]["links"]["r3-link3"][addr_type].split("/")[0],
        ]
        static_routes = {
            "r3": {
                "static_routes": [
                    {
                        "network": [NETWORK1_1[addr_type]],
                    }
                ]
            }
        }

        result = verify_bgp_rib(
            tgen, addr_type, "r3", static_routes, next_hop=next_hop_vrf
        )
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

        result = verify_rib(tgen, addr_type, "r3", static_routes, next_hop=next_hop_vrf)
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_locally_imported_routes_selected_as_bestpath_over_ebgp_imported_routes_p0(
    request,
):
    """
    Verify ECMP for imported routes from different VRFs.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)
    reset_config_on_routers(tgen)

    step(
        "Configure same static routes on R2 and R3 vrfs and redistribute in BGP "
        "for GREEN and RED vrf instances"
    )
    for dut, network in zip(
        ["r2", "r3"], [[NETWORK1_1, NETWORK1_2], [NETWORK1_1, NETWORK1_2]]
    ):
        for vrf_name, network_vrf in zip(["RED", "GREEN"], network):
            step("Configure static route for VRF : {} on {}".format(vrf_name, dut))
            for addr_type in ADDR_TYPES:
                static_routes = {
                    dut: {
                        "static_routes": [
                            {
                                "network": [network_vrf[addr_type]],
                                "next_hop": "blackhole",
                                "vrf": vrf_name,
                            }
                        ]
                    }
                }

                result = create_static_routes(tgen, static_routes)
                assert result is True, "Testcase {} :Failed \n Error: {}".format(
                    tc_name, result
                )

    for dut, as_num in zip(["r2", "r3"], ["2", "3"]):
        for vrf_name in ["RED", "GREEN"]:
            step("Redistribute static route on BGP VRF : {}".format(vrf_name))
            temp = {}
            for addr_type in ADDR_TYPES:
                temp.update(
                    {
                        addr_type: {
                            "unicast": {"redistribute": [{"redist_type": "static"}]}
                        }
                    }
                )

            redist_dict = {
                dut: {
                    "bgp": [
                        {"vrf": vrf_name, "local_as": as_num, "address_family": temp}
                    ]
                }
            }

            result = create_router_bgp(tgen, topo, redist_dict)
            assert result is True, "Testcase {} :Failed \n Error: {}".format(
                tc_name, result
            )

    step(
        "Verify that R2 and R3 has installed redistributed routes in default "
        "and RED vrfs and GREEN respectively:"
    )
    for dut, network in zip(
        ["r2", "r3"], [[NETWORK1_1, NETWORK1_2], [NETWORK1_1, NETWORK1_2]]
    ):
        for vrf_name, network_vrf in zip(["RED", "GREEN"], network):
            for addr_type in ADDR_TYPES:
                static_routes = {
                    dut: {
                        "static_routes": [
                            {
                                "network": [network_vrf[addr_type]],
                                "next_hop": "blackhole",
                                "vrf": vrf_name,
                            }
                        ]
                    }
                }
                result = verify_bgp_rib(tgen, addr_type, dut, static_routes)
                assert result is True, "Testcase {} : Failed \n Error {}".format(
                    tc_name, result
                )

    step("Import vrf RED's route in vrf GREEN on R3")
    temp = {}
    for addr_type in ADDR_TYPES:
        temp.update({addr_type: {"unicast": {"import": {"vrf": "RED"}}}})

    import_dict = {
        "r3": {"bgp": [{"vrf": "GREEN", "local_as": 3, "address_family": temp}]}
    }

    result = create_router_bgp(tgen, topo, import_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step(
        "Verify that locally imported routes are installed over eBGP imported"
        " routes from VRF RED into VRF GREEN"
    )
    for addr_type in ADDR_TYPES:
        static_routes = {
            "r3": {
                "static_routes": [
                    {
                        "network": [NETWORK1_2[addr_type]],
                        "next_hop": "blackhole",
                        "vrf": "GREEN",
                    }
                ]
            }
        }

        input_routes = {
            "r3": {
                addr_type: [
                    {
                        "network": NETWORK1_2[addr_type],
                        "bestpath": BESTPATH[addr_type],
                        "vrf": "GREEN",
                    }
                ]
            }
        }

        result = verify_bgp_bestpath(tgen, addr_type, input_routes)
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

        result = verify_rib(tgen, addr_type, "r3", static_routes)
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2024 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

import os
import re
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, get_topogen

pytestmark = [pytest.mark.bgpd]


def setup_module(mod):
    topodef = {"s1": ("r1", "r2", "r3")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_remote_as_auto():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]

    def _bgp_converge():
        output = json.loads(r1.vtysh_cmd("show bgp ipv4 unicast summary json"))
        expected = {
            "peers": {
                "192.168.1.2": {
                    "remoteAs": 65002,
                    "state": "Established",
                    "peerState": "OK",
                },
                "192.168.1.3": {
                    "remoteAs": 65003,
                    "state": "Established",
                    "peerState": "OK",
                },
            },
            "totalPeers": 2,
        }

        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't converge initial state"

    def _bgp_update_groups():
        actual = []
        output = json.loads(r1.vtysh_cmd("show bgp ipv4 unicast update-groups json"))
        expected = [
            {"subGroup": [{"adjListCount": 1, "peers": ["192.168.1.2"]}]},
            {"subGroup": [{"adjListCount": 1, "peers": ["192.168.1.3"]}]},
        ]

        # update-group's number can be random and it's not deterministic,
        # so we need to normalize the data a bit before checking.
        # We care here about the `peers` array only actually.
        for updgrp in output["default"].keys():
            actual.append(output["default"][updgrp])

        return topotest.json_cmp(actual, expected)

    test_func = functools.partial(
        _bgp_update_groups,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't see separate update-groups"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2023 by
# Adriano Marto Reis <adrianomarto@gmail.com>
#

import os
import sys
import json
import subprocess
from functools import partial
import pytest

from lib import topotest
from lib.topogen import Topogen, get_topogen
from lib.topolog import logger

from lib.common_config import verify_rib
from lib.ospf import verify_ospf_rib

pytestmark = pytest.mark.ospfd

"""
A large set of routers are connected to the same switch. Each router shares a
single network. All shared networks must be reachable from all routers.
"""

TOPOLOGY = """
       net1             net2                              netN
       ---              ---                               ---
        |                |            OSPF-passive         |
        |                |                                 |
    +---+---+        +---+---+                         +---+---+
    |       |        |       |                         |       |
    |  r1   |        |  r2   |          (...)          |  rN   |
    |       |        |       |                         |       |
    +---+---+        +---+---+                         +---+---+
        |                |           OSPF-active          |
        |                |                                |
        +----------------+--------------------------------+
                             switch


"""

N_ROUTERS = 8

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))


def build_topo(tgen):
    "Build the topology"

    # Create a single switch to connect all the routers
    switch = tgen.add_switch("switch")

    # Create routers
    for router_id in range(1, N_ROUTERS + 1):
        router = tgen.add_router(f"r{router_id}")
        switch.add_link(router)

        # The shared network needs to be connected to something
        dummy = tgen.add_switch(f"s{router_id}")
        dummy.add_link(router)


def setup_module(mod):
    logger.info("OSPF single switch:\n {}".format(TOPOLOGY))

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    # Starting Routers
    router_list = tgen.routers()

    for rname, router in router_list.items():
        logger.info("Loading router %s" % rname)
        router.load_frr_config(os.path.join(CWD, "{}_frr.conf".format(rname)))

    # Initialize all routers.
    tgen.start_router()


def teardown_module():
    "Tear-down the test environment"
    tgen = get_topogen()
    tgen.stop_topology()


def is_iproute2_json_supported():
    """
    Checks if the command 'ip -j route' is supported.
    """
    try:
        output = subprocess.run(
            ["ip", "-j", "route", "get", "0.0.0.0"], stdout=subprocess.PIPE
        ).stdout.decode()
        json.loads(output)
        return True
    except json.decoder.JSONDecodeError:
        return False


@pytest.mark.skipif(
    not is_iproute2_json_supported(), reason="'ip -j route' not supported"
)
def test_all_routes_advertised():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    # networks advertised by each router and the expected next hops
    networks = {
        "r1": ("10.0.1.0/24", "203.0.113.1"),
        "r2": ("10.0.2.0/24", "203.0.113.2"),
        "r3": ("10.0.3.0/24", "203.0.113.3"),
        "r4": ("10.0.4.0/24", "203.0.113.4"),
        "r5": ("10.0.5.0/24", "203.0.113.5"),
        "r6": ("10.0.6.0/24", "203.0.113.6"),
        "r7": ("10.0.7.0/24", "203.0.113.7"),
        "r8": ("10.0.8.0/24", "203.0.113.8"),
    }

    for router_orig in tgen.routers().keys():
        for router_dest, network in networks.items():
            if router_orig != router_dest:
                input_dict = {
                    router_orig: {
                        "static_routes": [
                            {
                                "network": network[0],
                            }
                        ]
                    }
                }
                result = verify_ospf_rib(
                    tgen, router_orig, input_dict, next_hop=network[1]
                )
                assert result is True, "Error: {}".format(result)
                result = verify_rib(
                    tgen, "ipv4", router_orig, input_dict, next_hop=network[1]
                )
                assert result is True, "Error: {}".format(result)

                check_route(router_orig, network[0], network[1])


def check_route(router_name, network, expected_nexthop):
    """
    Checks if the given network is present on the given router and has the
    expected next hop.
    """
    tgen = get_topogen()
    router = tgen.gears[router_name]

    expected_response = {
        network: [
            {
                "prefix": network,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": expected_nexthop,
                        "active": True,
                    },
                ],
            },
        ],
    }

    test_func = partial(
        topotest.router_json_cmp,
        router,
        f"show ip route {network} json",
        expected_response,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assert (
        result is None
    ), f"{router_name} (ospfd): no route {network} via {expected_nexthop}"

    address = network.split("/")[0]
    output = router.cmd(f"ip -j route get {address}")
    logger.info(output)
    routes = json.loads(output)
    assert (
        routes[0]["gateway"] == expected_nexthop
    ), f"{router_name} (kernel): no route {address} via {expected_nexthop}"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2020 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation, Inc.
# ("NetDEF") in this file.
#


"""OSPF Summarisation Functionality Automation."""
import os
import sys
import time
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen
from time import sleep

# Import topoJson from lib, to create topology and initial configuration
from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    reset_config_on_routers,
    verify_rib,
    create_static_routes,
    step,
)
from lib.topolog import logger
from lib.topojson import build_config_from_json
from lib.ospf import (
    verify_ospf_neighbor,
    verify_ospf_rib,
    create_router_ospf,
    verify_ospf_summary,
)

pytestmark = [pytest.mark.ospfd, pytest.mark.staticd]


# Global variables
topo = None

NETWORK = {
    "ipv4": [
        "11.0.20.1/32",
        "11.0.20.2/32",
        "11.0.20.3/32",
        "11.0.20.4/32",
        "11.0.20.5/32",
    ]
}
NETWORK2 = {
    "ipv4": [
        "12.0.20.1/32",
        "12.0.20.2/32",
        "12.0.20.3/32",
        "12.0.20.4/32",
        "12.0.20.5/32",
    ]
}
NETWORK3 = {
    "ipv4": [
        "13.0.20.1/32",
        "13.0.20.2/32",
        "13.0.20.3/32",
        "13.0.20.4/32",
        "13.0.20.5/32",
    ]
}
SUMMARY = {"ipv4": ["11.0.20.1/8", "12.0.0.0/8", "13.0.0.0/8", "11.0.0.0/8"]}
"""
TOPOOLOGY =
      Please view in a fixed-width font such as Courier.
      +---+  A1       +---+
      +R1 +------------+R2 |
      +-+-+-           +--++
        |  --        --  |
        |    -- A0 --    |
      A0|      ----      |
        |      ----      | A2
        |    --    --    |
        |  --        --  |
      +-+-+-            +-+-+
      +R0 +-------------+R3 |
      +---+     A3     +---+



TESTCASES =
1. OSPF summarisation with type7 LSAs.

"""


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """
    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/ospf_asbr_summary_type7_lsa.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    # Api call verify whether OSPF is converged
    ospf_covergence = verify_ospf_neighbor(tgen, topo)
    assert ospf_covergence is True, "setup_module :Failed \n Error  {}".format(
        ospf_covergence
    )

    logger.info("Running setup_module() done")


def teardown_module():
    """
    Teardown the pytest environment.

    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


def red_static(dut, config=True):
    """Local def for Redstribute static routes inside ospf."""
    global topo
    tgen = get_topogen()
    if config:
        ospf_red = {dut: {"ospf": {"redistribute": [{"redist_type": "static"}]}}}
    else:
        ospf_red = {
            dut: {"ospf": {"redistribute": [{"redist_type": "static", "delete": True}]}}
        }
    result = create_router_ospf(tgen, topo, ospf_red)
    assert result is True, "Testcase : Failed \n Error: {}".format(result)


def red_connected(dut, config=True):
    """Local def for Redstribute connected routes inside ospf."""
    global topo
    tgen = get_topogen()
    if config:
        ospf_red = {dut: {"ospf": {"redistribute": [{"redist_type": "connected"}]}}}
    else:
        ospf_red = {
            dut: {
                "ospf": {
                    "redistribute": [{"redist_type": "connected", "del_action": True}]
                }
            }
        }
    result = create_router_ospf(tgen, topo, ospf_red)
    assert result is True, "Testcase: Failed \n Error: {}".format(result)


# ##################################
# Test cases start here.
# ##################################


def test_ospf_type5_summary_tc44_p0(request):
    """OSPF summarisation with type7 LSAs"""

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Bring up the base config as per the topology")
    step("Configure area 1 as NSSA Area")

    reset_config_on_routers(tgen)

    dut = "r0"
    protocol = "ospf"

    red_static(dut)
    input_dict_static_rtes = {
        "r0": {
            "static_routes": [
                {"network": NETWORK["ipv4"], "next_hop": "blackhole"},
                {"network": NETWORK2["ipv4"], "next_hop": "blackhole"},
            ]
        }
    }
    result = create_static_routes(tgen, input_dict_static_rtes)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that routes are learnt on R1.")
    dut = "r1"

    result = verify_ospf_rib(tgen, dut, input_dict_static_rtes)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)
    result = verify_rib(tgen, "ipv4", dut, input_dict_static_rtes, protocol=protocol)
    assert (
        result is True
    ), "Testcase {} : Failed. Error: Routes is missing in RIB".format(tc_name)

    step("Configure External Route summary in R0 to summarise 5  routes to one route.")

    ospf_summ_r0 = {
        "r0": {
            "ospf": {
                "summary-address": [
                    {"prefix": SUMMARY["ipv4"][0].split("/")[0], "mask": "8"}
                ]
            }
        }
    }
    result = create_router_ospf(tgen, topo, ospf_summ_r0)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step(
        "Verify that external routes are summarised to configured summary "
        "address on R0 after 5 secs of delay timer expiry and only one "
        "route is sent to R1."
    )

    step("Configure summary & redistribute static/connected route with  metric type 2")

    input_dict_summary = {"r0": {"static_routes": [{"network": SUMMARY["ipv4"][3]}]}}
    dut = "r1"

    result = verify_ospf_rib(tgen, dut, input_dict_summary)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    result = verify_rib(tgen, "ipv4", dut, input_dict_summary, protocol=protocol)
    assert (
        result is True
    ), "Testcase {} : Failed. Error: Routes is missing in RIB".format(tc_name)

    step("Verify that show ip ospf summary should show the summaries.")
    input_dict = {
        SUMMARY["ipv4"][3]: {
            "summaryAddress": SUMMARY["ipv4"][3],
            "metricType": "E2",
            "metric": 20,
            "tag": 0,
            "externalRouteCount": 5,
        }
    }
    dut = "r0"
    result = verify_ospf_summary(tgen, topo, dut, input_dict)
    assert (
        result is True
    ), "Testcase {} : Failed. Error: Summary missing in OSPF DB".format(tc_name)

    step("Learn type 7 lsa from neighbours")

    dut = "r1"
    protocol = "ospf"

    red_static(dut)
    input_dict_static_rtes = {
        "r1": {
            "static_routes": [{"network": NETWORK3["ipv4"], "next_hop": "blackhole"}]
        }
    }
    result = create_static_routes(tgen, input_dict_static_rtes)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that routes are learnt on R0.")
    dut = "r0"

    result = verify_ospf_rib(tgen, dut, input_dict_static_rtes)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)
    result = verify_rib(tgen, "ipv4", dut, input_dict_static_rtes, protocol=protocol)
    assert (
        result is True
    ), "Testcase {} : Failed. Error: Routes is missing in RIB".format(tc_name)

    ospf_summ_r0 = {
        "r0": {
            "ospf": {
                "summary-address": [
                    {"prefix": SUMMARY["ipv4"][2].split("/")[0], "mask": "8"}
                ]
            }
        }
    }
    result = create_router_ospf(tgen, topo, ospf_summ_r0)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that type7 LSAs received from neighbor are not summarised.")
    input_dict = {
        "13.0.0.0/8": {
            "summaryAddress": "13.0.0.0/8",
            "metricType": "E2",
            "metric": 20,
            "tag": 0,
            "externalRouteCount": 0,
        }
    }
    dut = "r0"
    result = verify_ospf_summary(tgen, topo, dut, input_dict)
    assert (
        result is True
    ), "Testcase {} : Failed. Error: Summary missing in OSPF DB".format(tc_name)

    step("Verify that already originated summary is intact.")
    input_dict = {
        SUMMARY["ipv4"][3]: {
            "summaryAddress": SUMMARY["ipv4"][3],
            "metricType": "E2",
            "metric": 20,
            "tag": 0,
            "externalRouteCount": 5,
        }
    }
    dut = "r0"
    result = verify_ospf_summary(tgen, topo, dut, input_dict)
    assert (
        result is True
    ), "Testcase {} : Failed. Error: Summary missing in OSPF DB".format(tc_name)

    dut = "r1"
    aggr_timer = {"r1": {"ospf": {"aggr_timer": 6}}}
    result = create_router_ospf(tgen, topo, aggr_timer)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)
    ospf_summ_r0 = {
        "r0": {
            "ospf": {
                "summary-address": [
                    {"prefix": SUMMARY["ipv4"][2].split("/")[0], "mask": "8"}
                ]
            }
        }
    }
    result = create_router_ospf(tgen, topo, ospf_summ_r0)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step(
        "wait for 6+1 seconds as ospf aggregation start after 6 secs as "
        "per the above aggr_timer command"
    )
    sleep(7)
    dut = "r1"
    aggr_timer = {"r1": {"ospf": {"del_aggr_timer": 6}}}
    result = create_router_ospf(tgen, topo, aggr_timer)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_ospf_netns_vrf.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_ospf_netns_vrf.py: Test OSPF with Network Namespace VRFs.
"""

import os
import sys
from functools import partial
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.ospfd]


def build_topo(tgen):
    "Build function"

    # Create 3 routers
    for routern in range(1, 4):
        tgen.add_router("r{}".format(routern))

    # Create a empty network for router 1
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])

    # Create a empty network for router 2
    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])

    # Interconect router 1, 2 and 3
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    # Create empty netowrk for router3
    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["r3"])


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    # check for zebra capability
    for rname, router in router_list.items():
        if router.check_capability(TopoRouter.RD_ZEBRA, "--vrfwnetns") == False:
            return pytest.skip(
                "Skipping OSPF VRF NETNS feature. VRF NETNS backend not available on FRR"
            )

    if os.system("ip netns list") != 0:
        return pytest.skip(
            "Skipping OSPF VRF NETNS Test. NETNS not available on System"
        )

    logger.info("Testing with VRF Namespace support")

    for rname, router in router_list.items():
        # create VRF rx-ospf-cust1 and link rx-eth{0,1} to rx-ospf-cust1
        ns = "{}-ospf-cust1".format(rname)
        router.net.add_netns(ns)
        router.net.set_intf_netns(rname + "-eth0", ns, up=True)
        router.net.set_intf_netns(rname + "-eth1", ns, up=True)

        router.use_netns_vrf()
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_OSPF, os.path.join(CWD, "{}/ospfd.conf".format(rname))
        )

    # Initialize all routers.
    tgen.start_router()
    for router in router_list.values():
        if router.has_version("<", "4.0"):
            tgen.set_error("unsupported version")


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()

    # Move interfaces out of vrf namespace and delete the namespace
    router_list = tgen.routers()
    for rname, _ in router_list.items():
        tgen.net[rname].reset_intf_netns(rname + "-eth0")
        tgen.net[rname].reset_intf_netns(rname + "-eth1")
        tgen.net[rname].delete_netns(rname + "-ospf-cust1")
    tgen.stop_topology()


# Shared test function to validate expected output.
def compare_show_ip_route_vrf(rname, expected):
    """
    Calls 'show ip ospf vrf [rname]-ospf-cust1 route' for router `rname` and compare the obtained
    result with the expected output.
    """
    tgen = get_topogen()
    vrf_name = "{0}-ospf-cust1".format(rname)
    current = topotest.ip4_route_zebra(tgen.gears[rname], vrf_name)
    ret = topotest.difflines(
        current, expected, title1="Current output", title2="Expected output"
    )
    return ret


def test_ospf_convergence():
    "Test OSPF daemon convergence"
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    for rname, router in tgen.routers().items():
        logger.info('Waiting for router "%s" convergence', rname)

        # Load expected results from the command
        reffile = os.path.join(CWD, "{}/ospfroute.txt".format(rname))
        expected = open(reffile).read()

        # Run test function until we get an result. Wait at most 60 seconds.
        test_func = partial(
            topotest.router_output_cmp,
            router,
            "show ip ospf vrf {0}-ospf-cust1 route".format(rname),
            expected,
        )
        result, diff = topotest.run_and_expect(test_func, "", count=160, wait=0.5)
        assertmsg = "OSPF did not converge on {}:\n{}".format(rname, diff)
        assert result, assertmsg


def test_ospf_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info('Checking OSPF IPv4 kernel routes in "%s"', router.name)
        reffile = os.path.join(CWD, "{}/zebraroute.txt".format(router.name))
        expected = open(reffile).read()
        # Run test function until we get an result. Wait at most 60 seconds.
        test_func = partial(compare_show_ip_route_vrf, router.name, expected)
        result, diff = topotest.run_and_expect(test_func, "", count=140, wait=0.5)
        assertmsg = 'OSPF IPv4 route mismatch in router "{}": {}'.format(
            router.name, diff
        )
        assert result, assertmsg


def test_ospf_json():
    "Test 'show ip ospf json' output for coherency."
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    for rname, router in tgen.routers().items():
        logger.info(
            'Comparing router "%s" "show ip ospf vrf %s-ospf-cust1 json" output',
            router.name,
            router.name,
        )
        expected = {
            "{}-ospf-cust1".format(router.name): {
                "vrfName": "{}-ospf-cust1".format(router.name),
                "routerId": "10.0.255.{}".format(rname[1:]),
                "tosRoutesOnly": True,
                "rfc2328Conform": True,
                "spfScheduleDelayMsecs": 0,
                "holdtimeMinMsecs": 50,
                "holdtimeMaxMsecs": 5000,
                "lsaMinIntervalMsecs": 5000,
                "lsaMinArrivalMsecs": 1000,
                "writeMultiplier": 20,
                "refreshTimerMsecs": 10000,
                "asbrRouter": "injectingExternalRoutingInformation",
                "attachedAreaCounter": 1,
                "areas": {},
            }
        }
        # Area specific additional checks
        if router.name == "r1" or router.name == "r2" or router.name == "r3":
            expected["{}-ospf-cust1".format(router.name)]["areas"]["0.0.0.0"] = {
                "areaIfActiveCounter": 2,
                "areaIfTotalCounter": 2,
                "authentication": "authenticationNone",
                "backbone": True,
                "lsaAsbrNumber": 0,
                "lsaNetworkNumber": 1,
                "lsaNssaNumber": 0,
                "lsaNumber": 4,
                "lsaOpaqueAreaNumber": 0,
                "lsaOpaqueLinkNumber": 0,
                "lsaRouterNumber": 3,
                "lsaSummaryNumber": 0,
                "nbrFullAdjacentCounter": 2,
            }

        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show ip ospf vrf {0}-ospf-cust1 json".format(rname),
            expected,
        )
        _, diff = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = '"{}" JSON output mismatches'.format(rname)
        assert diff is None, assertmsg


def test_ospf_link_down():
    "Test OSPF convergence after a link goes down"
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    # Simulate a network down event on router3 switch3 interface.
    router3 = tgen.gears["r3"]
    topotest.interface_set_status(
        router3, "r3-eth0", ifaceaction=False, vrf_name="r3-ospf-cust1"
    )

    # Expect convergence on all routers
    for rname, router in tgen.routers().items():
        logger.info('Waiting for router "%s" convergence after link failure', rname)
        # Load expected results from the command
        reffile = os.path.join(CWD, "{}/ospfroute_down.txt".format(rname))
        expected = open(reffile).read()

        # Run test function until we get an result. Wait at most 60 seconds.
        test_func = partial(
            topotest.router_output_cmp,
            router,
            "show ip ospf vrf {0}-ospf-cust1 route".format(rname),
            expected,
        )
        result, diff = topotest.run_and_expect(test_func, "", count=140, wait=0.5)
        assertmsg = "OSPF did not converge on {}:\n{}".format(rname, diff)
        assert result, assertmsg


def test_ospf_link_down_kernel_route():
    "Test OSPF kernel route installation"
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    rlist = tgen.routers().values()
    for router in rlist:
        logger.info(
            'Checking OSPF IPv4 kernel routes in "%s" after link down', router.name
        )

        str = "{0}-ospf-cust1".format(router.name)
        reffile = os.path.join(CWD, "{}/zebraroutedown.txt".format(router.name))
        expected = open(reffile).read()
        # Run test function until we get an result. Wait at most 60 seconds.
        test_func = partial(compare_show_ip_route_vrf, router.name, expected)
        result, diff = topotest.run_and_expect(test_func, "", count=140, wait=0.5)
        assertmsg = (
            'OSPF IPv4 route mismatch in router "{}" after link down: {}'.format(
                router.name, diff
            )
        )
        assert result, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_sender-as-path-loop-detection.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2019 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#

"""
Test if neighbor <neighbor> sender-as-path-loop-detection
command works as expeced.
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen


pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    for routern in range(1, 4):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_sender_as_path_loop_detection():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]

    def _bgp_converge():
        output = json.loads(r2.vtysh_cmd("show ip bgp neighbor 192.168.255.2 json"))
        expected = {
            "192.168.255.2": {
                "bgpState": "Established",
                "addressFamilyInfo": {"ipv4Unicast": {"acceptedPrefixCounter": 3}},
            }
        }
        return topotest.json_cmp(output, expected)

    def _bgp_has_route_from_r1():
        output = json.loads(r2.vtysh_cmd("show ip bgp 172.16.255.253/32 json"))
        expected = {
            "paths": [
                {
                    "aspath": {
                        "segments": [{"type": "as-sequence", "list": [65001, 65003]}],
                        "length": 2,
                    }
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    def _bgp_suppress_route_to_r1():
        output = json.loads(
            r2.vtysh_cmd("show ip bgp neighbor 192.168.255.2 advertised-routes json")
        )
        expected = {"totalPrefixCounter": 0}
        return topotest.json_cmp(output, expected)

    def _bgp_suppress_route_to_r3():
        output = json.loads(
            r2.vtysh_cmd("show ip bgp neighbor 192.168.254.2 advertised-routes json")
        )
        expected = {"totalPrefixCounter": 2}
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_converge)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Failed bgp to convergence"

    test_func = functools.partial(_bgp_has_route_from_r1)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Failed to see a route from r1"

    test_func = functools.partial(_bgp_suppress_route_to_r3)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Route 172.16.255.253/32 should not be sent to r3 from r2"

    test_func = functools.partial(_bgp_suppress_route_to_r1)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Routes should not be sent to r1 from r2"


def test_remove_loop_detection_on_one_peer():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]

    def _bgp_reset_route_to_r1():
        output = json.loads(
            r2.vtysh_cmd("show ip bgp neighbor 192.168.255.2 advertised-routes json")
        )
        expected = {"totalPrefixCounter": 3}
        return topotest.json_cmp(output, expected)

    r2.vtysh_cmd(
        """
        configure terminal
         router bgp 65002
          no neighbor 192.168.255.2 sender-as-path-loop-detection
        """
    )

    r2.vtysh_cmd(
        """
        clear bgp *
        """
    )
    test_func = functools.partial(_bgp_reset_route_to_r1)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Failed bgp to reset route"


def test_loop_detection_on_peer_group():
    tgen = get_topogen()

    r2 = tgen.gears["r2"]

    def _bgp_suppress_route_to_r1():
        output = json.loads(
            r2.vtysh_cmd("show ip bgp neighbor 192.168.255.2 advertised-routes json")
        )
        expected = {"totalPrefixCounter": 0}
        return topotest.json_cmp(output, expected)

    def _bgp_suppress_route_to_r3():
        output = json.loads(
            r2.vtysh_cmd("show ip bgp neighbor 192.168.254.2 advertised-routes json")
        )
        expected = {"totalPrefixCounter": 2}
        return topotest.json_cmp(output, expected)

    r2.vtysh_cmd(
        """
        configure terminal
         router bgp 65002
          neighbor loop_group peer-group
          neighbor 192.168.255.2 peer-group loop_group
          neighbor loop_group sender-as-path-loop-detection
        """
    )

    r2.vtysh_cmd(
        """
        clear bgp *
        """
    )

    test_func = functools.partial(_bgp_suppress_route_to_r3)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Route 172.16.255.253/32 should not be sent to r3 from r2"

    test_func = functools.partial(_bgp_suppress_route_to_r1)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Routes should not be sent to r1 from r2"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
#
# April 14 2024, Christian Hopps <chopps@labn.net>
#
# Copyright (c) 2024, LabN Consulting, L.L.C.
#
"""Common CLI execute argument."""


def add_launch_args(add_func):

    add_func("--gdb", metavar="NODE-LIST", help="comma-sep list of hosts to run gdb on")
    add_func(
        "--gdb-breakpoints",
        metavar="BREAKPOINT-LIST",
        help="comma-sep list of breakpoints to set",
    )
    add_func(
        "--gdb-use-emacs",
        action="store_true",
        help="Use emacsclient to run gdb instead of a shell",
    )

    add_func(
        "--host",
        action="store_true",
        help="no isolation for top namespace, bridges exposed to default namespace",
    )
    add_func(
        "--pcap",
        metavar="TARGET-LIST",
        help="comma-sep list of capture targets (NETWORK or NODE:IFNAME) or 'all'",
    )
    add_func(
        "--shell", metavar="NODE-LIST", help="comma-sep list of nodes to open shells on"
    )
    add_func(
        "--stderr",
        metavar="NODE-LIST",
        help="comma-sep list of nodes to open windows viewing stderr",
    )
    add_func(
        "--stdout",
        metavar="NODE-LIST",
        help="comma-sep list of nodes to open windows viewing stdout",
    )


def add_testing_args(add_func):
    add_func(
        "--cli-on-error",
        action="store_true",
        help="CLI on test failure",
    )

    add_func(
        "--coverage",
        action="store_true",
        help="Enable coverage gathering if supported",
    )

    add_func(
        "--cov-build-dir",
        help="Specify the build dir for locating coverage data files",
    )

    add_launch_args(add_func)

    add_func(
        "--pause",
        action="store_true",
        help="Pause after each test",
    )
    add_func(
        "--pause-at-end",
        action="store_true",
        help="Pause before taking munet down",
    )
    add_func(
        "--pause-on-error",
        action="store_true",
        help="Pause after (disables default when --shell or -vtysh given)",
    )
    add_func(
        "--no-pause-on-error",
        dest="pause_on_error",
        action="store_false",
        help="Do not pause after (disables default when --shell or -vtysh given)",
    )



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2020 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test BGP VRF Lite:

1. Verify BGP best path selection algorithm works fine when
routes are imported from ISR to default vrf and vice versa.
"""

import os
import sys
import time
import pytest
import platform

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# Required to instantiate the topology builder class.

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen
from lib.topotest import version_cmp

from lib.common_config import (
    start_topology,
    write_test_header,
    check_address_types,
    write_test_footer,
    step,
    create_route_maps,
    create_prefix_lists,
    check_router_status,
    get_frr_ipv6_linklocal,
    shutdown_bringup_interface,
)

from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_bgp_rib,
    verify_best_path_as_per_bgp_attribute,
)
from lib.topojson import build_config_from_json


pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]

# Global variables
NETWORK1_1 = {"ipv4": "11.11.11.1/32", "ipv6": "11:11::1/128"}
NETWORK1_2 = {"ipv4": "11.11.11.11/32", "ipv6": "11:11::11/128"}
NETWORK1_3 = {"ipv4": "10.10.10.10/32", "ipv6": "10:10::10/128"}
NETWORK1_4 = {"ipv4": "10.10.10.100/32", "ipv6": "10:10::100/128"}

NETWORK2_1 = {"ipv4": "22.22.22.2/32", "ipv6": "22:22::2/128"}
NETWORK2_2 = {"ipv4": "22.22.22.22/32", "ipv6": "22:22::22/128"}
NETWORK2_3 = {"ipv4": "20.20.20.20/32", "ipv6": "20:20::20/128"}
NETWORK2_4 = {"ipv4": "20.20.20.200/32", "ipv6": "20:20::200/128"}

NETWORK3_1 = {"ipv4": "30.30.30.3/32", "ipv6": "30:30::3/128"}
NETWORK3_2 = {"ipv4": "30.30.30.30/32", "ipv6": "30:30::30/128"}
NETWORK3_3 = {"ipv4": "50.50.50.5/32", "ipv6": "50:50::5/128"}
NETWORK3_4 = {"ipv4": "50.50.50.50/32", "ipv6": "50:50::50/128"}

NETWORK4_1 = {"ipv4": "40.40.40.4/32", "ipv6": "40:40::4/128"}
NETWORK4_2 = {"ipv4": "40.40.40.40/32", "ipv6": "40:40::40/128"}
NETWORK4_3 = {"ipv4": "50.50.50.5/32", "ipv6": "50:50::5/128"}
NETWORK4_4 = {"ipv4": "50.50.50.50/32", "ipv6": "50:50::50/128"}
NEXT_HOP_IP = {"ipv4": "Null0", "ipv6": "Null0"}
LOOPBACK_1 = {
    "ipv4": "10.0.0.7/24",
    "ipv6": "fd00:0:0:1::7/64",
    "ipv4_mask": "255.255.255.0",
    "ipv6_mask": None,
}
LOOPBACK_2 = {
    "ipv4": "10.0.0.16/24",
    "ipv6": "fd00:0:0:3::5/64",
    "ipv4_mask": "255.255.255.0",
    "ipv6_mask": None,
}
PREFERRED_NEXT_HOP = "global"


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_vrf_lite_best_path_topo1.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Run these tests for kernel version 4.19 or above
    if version_cmp(platform.release(), "4.19") < 0:
        error_msg = (
            "BGP vrf dynamic route leak tests will not run "
            '(have kernel "{}", but it requires >= 4.19)'.format(platform.release())
        )
        pytest.skip(error_msg)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    global BGP_CONVERGENCE
    global ADDR_TYPES
    ADDR_TYPES = check_address_types()

    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "setup_module : Failed \n Error: {}".format(
        BGP_CONVERGENCE
    )

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
#
#   Testcases
#
#####################################################


def disable_route_map_to_prefer_global_next_hop(tgen, topo):
    """
    This API is to remove prefer global route-map applied on neighbors

    Parameter:
    ----------
    * `tgen` : Topogen object
    * `topo` : Input JSON data

    Returns:
    --------
    True/errormsg

    """

    logger.info("Remove prefer-global rmap applied on neighbors")
    input_dict = {
        "r1": {
            "bgp": [
                {
                    "local_as": "100",
                    "vrf": "ISR",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r2": {
                                        "dest_link": {
                                            "r1-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
                {
                    "local_as": "100",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r3": {
                                        "dest_link": {
                                            "r1-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
                {
                    "local_as": "100",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r4": {
                                        "dest_link": {
                                            "r1-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            ]
        },
        "r2": {
            "bgp": [
                {
                    "local_as": "100",
                    "vrf": "ISR",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r2-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
                {
                    "local_as": "100",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r3": {
                                        "dest_link": {
                                            "r2-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
                {
                    "local_as": "100",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r4": {
                                        "dest_link": {
                                            "r2-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            ]
        },
        "r3": {
            "bgp": [
                {
                    "local_as": "300",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r3-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
                {
                    "local_as": "300",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r2": {
                                        "dest_link": {
                                            "r3-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            ]
        },
        "r4": {
            "bgp": [
                {
                    "local_as": "400",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r1": {
                                        "dest_link": {
                                            "r4-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
                {
                    "local_as": "400",
                    "address_family": {
                        "ipv6": {
                            "unicast": {
                                "neighbor": {
                                    "r2": {
                                        "dest_link": {
                                            "r4-link1": {
                                                "route_maps": [
                                                    {
                                                        "name": "rmap_global",
                                                        "direction": "in",
                                                        "delete": True,
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            ]
        },
    }

    result = create_router_bgp(tgen, topo, input_dict)
    assert result is True, "Testcase :Failed \n Error: {}".format(result)

    return True


def test_bgp_best_path_with_dynamic_import_p0(request):
    """
    1.5.6. Verify BGP best path selection algorithm works fine when
    routes are imported from ISR to default vrf and vice versa.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)
    build_config_from_json(tgen, topo)

    if tgen.routers_have_failure():
        check_router_status(tgen)

    for addr_type in ADDR_TYPES:
        step(
            "Redistribute configured static routes into BGP process" " on R1/R2 and R3"
        )

        input_dict_1 = {}
        DUT = ["r1", "r2", "r3", "r4"]
        VRFS = ["ISR", "ISR", "default", "default"]
        AS_NUM = [100, 100, 300, 400]

        for dut, vrf, as_num in zip(DUT, VRFS, AS_NUM):
            temp = {dut: {"bgp": []}}
            input_dict_1.update(temp)

            temp[dut]["bgp"].append(
                {
                    "local_as": as_num,
                    "vrf": vrf,
                    "address_family": {
                        addr_type: {
                            "unicast": {"redistribute": [{"redist_type": "static"}]}
                        }
                    },
                }
            )

        result = create_router_bgp(tgen, topo, input_dict_1)
        assert result is True, "Testcase {} :Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        step("Import from default vrf into vrf ISR on R1 and R2 as below")

        input_dict_vrf = {}
        DUT = ["r1", "r2"]
        VRFS = ["ISR", "ISR"]
        AS_NUM = [100, 100]

        for dut, vrf, as_num in zip(DUT, VRFS, AS_NUM):
            temp = {dut: {"bgp": []}}
            input_dict_vrf.update(temp)

            temp[dut]["bgp"].append(
                {
                    "local_as": as_num,
                    "vrf": vrf,
                    "address_family": {
                        addr_type: {"unicast": {"import": {"vrf": "default"}}}
                    },
                }
            )

        result = create_router_bgp(tgen, topo, input_dict_vrf)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        input_dict_default = {}
        DUT = ["r1", "r2"]
        VRFS = ["default", "default"]
        AS_NUM = [100, 100]

        for dut, vrf, as_num in zip(DUT, VRFS, AS_NUM):
            temp = {dut: {"bgp": []}}
            input_dict_default.update(temp)

            temp[dut]["bgp"].append(
                {
                    "local_as": as_num,
                    "vrf": vrf,
                    "address_family": {
                        addr_type: {"unicast": {"import": {"vrf": "ISR"}}}
                    },
                }
            )

        result = create_router_bgp(tgen, topo, input_dict_default)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step(
        "Verify ECMP/Next-hop/Imported routes Vs Locally originated "
        "routes/eBGP routes vs iBGP routes --already covered in almost"
        " all tests"
    )

    for addr_type in ADDR_TYPES:
        step("Verify Pre-emption")

        input_routes_r3 = {
            "r3": {"static_routes": [{"network": [NETWORK3_3[addr_type]]}]}
        }

        intf_r3_r1 = topo["routers"]["r3"]["links"]["r1-link1"]["interface"]
        intf_r4_r1 = topo["routers"]["r4"]["links"]["r1-link1"]["interface"]

        if addr_type == "ipv6" and "link_local" in PREFERRED_NEXT_HOP:
            nh_r3_r1 = get_frr_ipv6_linklocal(tgen, "r3", intf=intf_r3_r1)
            nh_r4_r1 = get_frr_ipv6_linklocal(tgen, "r4", intf=intf_r4_r1)
        else:
            nh_r3_r1 = topo["routers"]["r3"]["links"]["r1-link1"][addr_type].split("/")[
                0
            ]
            nh_r4_r1 = topo["routers"]["r4"]["links"]["r1-link1"][addr_type].split("/")[
                0
            ]

        result = verify_bgp_rib(
            tgen, addr_type, "r1", input_routes_r3, next_hop=[nh_r4_r1]
        )
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

    step("Shutdown interface connected to r1 from r4:")
    shutdown_bringup_interface(tgen, "r4", intf_r4_r1, False)

    for addr_type in ADDR_TYPES:
        input_routes_r3 = {
            "r3": {"static_routes": [{"network": [NETWORK3_3[addr_type]]}]}
        }

        intf_r3_r1 = topo["routers"]["r3"]["links"]["r1-link1"]["interface"]
        intf_r4_r1 = topo["routers"]["r4"]["links"]["r1-link1"]["interface"]

        if addr_type == "ipv6" and "link_local" in PREFERRED_NEXT_HOP:
            nh_r3_r1 = get_frr_ipv6_linklocal(tgen, "r3", intf=intf_r3_r1)
            nh_r4_r1 = get_frr_ipv6_linklocal(tgen, "r4", intf=intf_r4_r1)
        else:
            nh_r3_r1 = topo["routers"]["r3"]["links"]["r1-link1"][addr_type].split("/")[
                0
            ]
            nh_r4_r1 = topo["routers"]["r4"]["links"]["r1-link1"][addr_type].split("/")[
                0
            ]

        step("Verify next-hop is changed")
        result = verify_bgp_rib(
            tgen, addr_type, "r1", input_routes_r3, next_hop=[nh_r3_r1]
        )
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

    step("Bringup interface connected to r1 from r4:")
    shutdown_bringup_interface(tgen, "r4", intf_r4_r1, True)

    for addr_type in ADDR_TYPES:
        input_routes_r3 = {
            "r3": {"static_routes": [{"network": [NETWORK3_3[addr_type]]}]}
        }

        intf_r3_r1 = topo["routers"]["r3"]["links"]["r1-link1"]["interface"]
        intf_r4_r1 = topo["routers"]["r4"]["links"]["r1-link1"]["interface"]

        if addr_type == "ipv6" and "link_local" in PREFERRED_NEXT_HOP:
            nh_r3_r1 = get_frr_ipv6_linklocal(tgen, "r3", intf=intf_r3_r1)
            nh_r4_r1 = get_frr_ipv6_linklocal(tgen, "r4", intf=intf_r4_r1)
        else:
            nh_r3_r1 = topo["routers"]["r3"]["links"]["r1-link1"][addr_type].split("/")[
                0
            ]
            nh_r4_r1 = topo["routers"]["r4"]["links"]["r1-link1"][addr_type].split("/")[
                0
            ]

        step("Verify next-hop is not chnaged aftr shutdown:")
        result = verify_bgp_rib(
            tgen, addr_type, "r1", input_routes_r3, next_hop=[nh_r3_r1]
        )
        assert result is True, "Testcase {} : Failed \n Error {}".format(
            tc_name, result
        )

    step("Active-Standby scenario(as-path prepend and Local pref)")

    for addr_type in ADDR_TYPES:
        step("Create prefix-list")

        input_dict_pf = {
            "r1": {
                "prefix_lists": {
                    addr_type: {
                        "pf_ls_{}".format(addr_type): [
                            {
                                "seqid": 10,
                                "network": NETWORK3_4[addr_type],
                                "action": "permit",
                            }
                        ]
                    }
                }
            }
        }
        result = create_prefix_lists(tgen, input_dict_pf)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        step("Create route-map to match prefix-list and set localpref 500")

        input_dict_rm = {
            "r1": {
                "route_maps": {
                    "rmap_PATH1_{}".format(addr_type): [
                        {
                            "action": "permit",
                            "seq_id": 10,
                            "match": {
                                addr_type: {
                                    "prefix_lists": "pf_ls_{}".format(addr_type)
                                }
                            },
                            "set": {"locPrf": 500},
                        }
                    ]
                }
            }
        }

        result = create_route_maps(tgen, input_dict_rm)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("Create route-map to match prefix-list and set localpref 600")

        input_dict_rm = {
            "r1": {
                "route_maps": {
                    "rmap_PATH2_{}".format(addr_type): [
                        {
                            "action": "permit",
                            "seq_id": 20,
                            "match": {
                                addr_type: {
                                    "prefix_lists": "pf_ls_{}".format(addr_type)
                                }
                            },
                            "set": {"locPrf": 600},
                        }
                    ]
                }
            }
        }

        result = create_route_maps(tgen, input_dict_rm)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        input_dict_rma = {
            "r1": {
                "bgp": [
                    {
                        "local_as": "100",
                        "address_family": {
                            addr_type: {
                                "unicast": {
                                    "neighbor": {
                                        "r3": {
                                            "dest_link": {
                                                "r1-link1": {
                                                    "route_maps": [
                                                        {
                                                            "name": "rmap_PATH1_{}".format(
                                                                addr_type
                                                            ),
                                                            "direction": "in",
                                                        }
                                                    ]
                                                }
                                            }
                                        },
                                        "r4": {
                                            "dest_link": {
                                                "r1-link1": {
                                                    "route_maps": [
                                                        {
                                                            "name": "rmap_PATH2_{}".format(
                                                                addr_type
                                                            ),
                                                            "direction": "in",
                                                        }
                                                    ]
                                                }
                                            }
                                        },
                                    }
                                }
                            }
                        },
                    }
                ]
            }
        }

        result = create_router_bgp(tgen, topo, input_dict_rma)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    dut = "r1"
    attribute = "locPrf"

    for addr_type in ADDR_TYPES:
        step("Verify bestpath is installed as per highest localpref")

        input_routes_r3 = {
            "r3": {"static_routes": [{"network": [NETWORK3_4[addr_type]]}]}
        }

        result = verify_best_path_as_per_bgp_attribute(
            tgen, addr_type, dut, input_routes_r3, attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        step("Create route-map to match prefix-list and set localpref 700")

        input_dict_rm = {
            "r1": {
                "route_maps": {
                    "rmap_PATH1_{}".format(addr_type): [
                        {
                            "action": "permit",
                            "seq_id": 10,
                            "match": {
                                addr_type: {
                                    "prefix_lists": "pf_ls_{}".format(addr_type)
                                }
                            },
                            "set": {"locPrf": 700},
                        }
                    ]
                }
            }
        }

        result = create_route_maps(tgen, input_dict_rm)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        step("Verify bestpath is changed as per highest localpref")

        input_routes_r3 = {
            "r3": {"static_routes": [{"network": [NETWORK3_4[addr_type]]}]}
        }

        result = verify_best_path_as_per_bgp_attribute(
            tgen, addr_type, dut, input_routes_r3, attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        step("Create route-map to match prefix-list and set as-path prepend")

        input_dict_rm = {
            "r1": {
                "route_maps": {
                    "rmap_PATH2_{}".format(addr_type): [
                        {
                            "action": "permit",
                            "seq_id": 20,
                            "match": {
                                addr_type: {
                                    "prefix_lists": "pf_ls_{}".format(addr_type)
                                }
                            },
                            "set": {
                                "localpref": 700,
                                "path": {"as_num": "111", "as_action": "prepend"},
                            },
                        }
                    ]
                }
            }
        }

        result = create_route_maps(tgen, input_dict_rm)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    attribute = "path"

    for addr_type in ADDR_TYPES:
        step("Verify bestpath is changed as per shortest as-path")

        input_routes_r3 = {
            "r3": {"static_routes": [{"network": [NETWORK3_4[addr_type]]}]}
        }

        result = verify_best_path_as_per_bgp_attribute(
            tgen, addr_type, dut, input_routes_r3, attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: GPL-2.0-or-later
# FRR CLI preprocessor (DEFPY)
#
# Copyright (C) 2017  David Lamparter for NetDEF, Inc.

import clippy, traceback, sys, os
from collections import OrderedDict
from functools import reduce
from pprint import pprint
from string import Template
from io import StringIO

# the various handlers generate output C code for a particular type of
# CLI token, choosing the most useful output C type.


class RenderHandler(object):
    def __init__(self, token):
        pass

    def combine(self, other):
        if type(self) == type(other):
            return other
        return StringHandler(None)

    deref = ""
    drop_str = False
    canfail = True
    canassert = False


class StringHandler(RenderHandler):
    argtype = "const char *"
    decl = Template("const char *$varname = NULL;")
    code = Template(
        "$varname = (argv[_i]->type == WORD_TKN) ? argv[_i]->text : argv[_i]->arg;"
    )
    drop_str = True
    canfail = False
    canassert = True


class LongHandler(RenderHandler):
    argtype = "long"
    decl = Template("long $varname = 0;")
    code = Template(
        """\
char *_end;
$varname = strtol(argv[_i]->arg, &_end, 10);
_fail = (_end == argv[_i]->arg) || (*_end != '\\0');"""
    )


class AsDotHandler(RenderHandler):
    argtype = "as_t"
    decl = Template("as_t $varname = 0;")
    code = Template("_fail = !asn_str2asn(argv[_i]->arg, &$varname);")


# A.B.C.D/M (prefix_ipv4) and
# X:X::X:X/M (prefix_ipv6) are "compatible" and can merge into a
# struct prefix:


class PrefixBase(RenderHandler):
    def combine(self, other):
        if type(self) == type(other):
            return other
        if isinstance(other, PrefixBase):
            return PrefixGenHandler(None)
        return StringHandler(None)

    deref = "&"


class Prefix4Handler(PrefixBase):
    argtype = "const struct prefix_ipv4 *"
    decl = Template("struct prefix_ipv4 $varname = { };")
    code = Template("_fail = !str2prefix_ipv4(argv[_i]->arg, &$varname);")


class Prefix6Handler(PrefixBase):
    argtype = "const struct prefix_ipv6 *"
    decl = Template("struct prefix_ipv6 $varname = { };")
    code = Template("_fail = !str2prefix_ipv6(argv[_i]->arg, &$varname);")


class PrefixEthHandler(PrefixBase):
    argtype = "struct prefix_eth *"
    decl = Template("struct prefix_eth $varname = { };")
    code = Template("_fail = !str2prefix_eth(argv[_i]->arg, &$varname);")


class PrefixGenHandler(PrefixBase):
    argtype = "const struct prefix *"
    decl = Template("struct prefix $varname = { };")
    code = Template("_fail = !str2prefix(argv[_i]->arg, &$varname);")


# same for IP addresses.  result is union sockunion.
class IPBase(RenderHandler):
    def combine(self, other):
        if type(self) == type(other):
            return other
        if type(other) in [IP4Handler, IP6Handler, IPGenHandler]:
            return IPGenHandler(None)
        return StringHandler(None)


class IP4Handler(IPBase):
    argtype = "struct in_addr"
    decl = Template("struct in_addr $varname = { INADDR_ANY };")
    code = Template("_fail = !inet_aton(argv[_i]->arg, &$varname);")


class IP6Handler(IPBase):
    argtype = "struct in6_addr"
    decl = Template("struct in6_addr $varname = {};")
    code = Template("_fail = !inet_pton(AF_INET6, argv[_i]->arg, &$varname);")


class IPGenHandler(IPBase):
    argtype = "const union sockunion *"
    decl = Template(
        """union sockunion s__$varname = { .sa.sa_family = AF_UNSPEC }, *$varname = NULL;"""
    )
    code = Template(
        """\
if (argv[_i]->text[0] == 'X') {
	s__$varname.sa.sa_family = AF_INET6;
	_fail = !inet_pton(AF_INET6, argv[_i]->arg, &s__$varname.sin6.sin6_addr);
	$varname = &s__$varname;
} else {
	s__$varname.sa.sa_family = AF_INET;
	_fail = !inet_aton(argv[_i]->arg, &s__$varname.sin.sin_addr);
	$varname = &s__$varname;
}"""
    )
    canassert = True


def mix_handlers(handlers):
    def combine(a, b):
        if a is None:
            return b
        return a.combine(b)

    return reduce(combine, handlers, None)


handlers = {
    "WORD_TKN": StringHandler,
    "VARIABLE_TKN": StringHandler,
    "RANGE_TKN": LongHandler,
    "IPV4_TKN": IP4Handler,
    "IPV4_PREFIX_TKN": Prefix4Handler,
    "IPV6_TKN": IP6Handler,
    "IPV6_PREFIX_TKN": Prefix6Handler,
    "MAC_TKN": PrefixEthHandler,
    "MAC_PREFIX_TKN": PrefixEthHandler,
    "ASNUM_TKN": AsDotHandler,
}

# core template invoked for each occurence of DEFPY.
#
# the "#if $..." bits are there to keep this template unified into one
# common form, without requiring a more advanced template engine (e.g.
# jinja2)
templ = Template(
    """$cond_begin/* $fnname => "$cmddef" */
DEFUN_CMD_FUNC_DECL($fnname)
#define funcdecl_$fnname static int ${fnname}_magic(\\
	const struct cmd_element *self __attribute__ ((unused)),\\
	struct vty *vty __attribute__ ((unused)),\\
	int argc __attribute__ ((unused)),\\
	struct cmd_token *argv[] __attribute__ ((unused))$argdefs)
funcdecl_$fnname;
DEFUN_CMD_FUNC_TEXT($fnname)
{
#if $nonempty /* anything to parse? */
	int _i;
#if $canfail /* anything that can fail? */
	unsigned _fail = 0, _failcnt = 0;
#endif
$argdecls
	for (_i = 0; _i < argc; _i++) {
		if (!argv[_i]->varname)
			continue;
#if $canfail /* anything that can fail? */
		_fail = 0;
#endif
$argblocks
#if $canfail /* anything that can fail? */
		if (_fail)
			vty_out (vty, "%% invalid input for %s: %s\\n",
				   argv[_i]->varname, argv[_i]->arg);
		_failcnt += _fail;
#endif
	}
#if $canfail /* anything that can fail? */
	if (_failcnt)
		return CMD_WARNING;
#endif
#endif
$argassert
	return ${fnname}_magic(self, vty, argc, argv$arglist);
}
$cond_end
"""
)

# invoked for each named parameter
argblock = Template(
    """
		if (!strcmp(argv[_i]->varname, \"$varname\")) {$strblock
			$code
		}"""
)


def get_always_args(token, always_args, args=[], stack=[]):
    if token in stack:
        return
    if token.type == "END_TKN":
        for arg in list(always_args):
            if arg not in args:
                always_args.remove(arg)
        return

    stack = stack + [token]
    if token.type in handlers and token.varname is not None:
        args = args + [token.varname]
    for nexttkn in token.next():
        get_always_args(nexttkn, always_args, args, stack)


class Macros(dict):
    def __init__(self):
        super().__init__()
        self._loc = {}

    def load(self, filename):
        filedata = clippy.parse(filename)
        for entry in filedata["data"]:
            if entry["type"] != "PREPROC":
                continue
            self.load_preproc(filename, entry)

    def setup(self, key, val, where="built-in"):
        self[key] = val
        self._loc[key] = (where, 0)

    def load_preproc(self, filename, entry):
        ppdir = entry["line"].lstrip().split(None, 1)
        if ppdir[0] != "define" or len(ppdir) != 2:
            return
        ppdef = ppdir[1].split(None, 1)
        name = ppdef[0]
        if "(" in name:
            return
        val = ppdef[1] if len(ppdef) == 2 else ""

        val = val.strip(" \t\n\\")
        if self.get(name, val) != val:
            sys.stderr.write(
                "%s:%d: warning: macro %s redefined!\n"
                % (
                    filename,
                    entry["lineno"],
                    name,
                )
            )
            sys.stderr.write(
                "%s:%d: note: previously defined here\n"
                % (
                    self._loc[name][0],
                    self._loc[name][1],
                )
            )
        else:
            self[name] = val
            self._loc[name] = (filename, entry["lineno"])


def process_file(fn, ofd, dumpfd, all_defun, macros):
    errors = 0
    filedata = clippy.parse(fn)

    cond_stack = []

    for entry in filedata["data"]:
        if entry["type"] == "PREPROC":
            line = entry["line"].lstrip()
            tokens = line.split(maxsplit=1)
            line = "#" + line + "\n"

            if not tokens:
                continue

            if tokens[0] in ["if", "ifdef", "ifndef"]:
                cond_stack.append(line)
            elif tokens[0] in ["elif", "else"]:
                prev_line = cond_stack.pop(-1)
                cond_stack.append(prev_line + line)
            elif tokens[0] in ["endif"]:
                cond_stack.pop(-1)
            elif tokens[0] in ["define"]:
                if not cond_stack:
                    macros.load_preproc(fn, entry)
                elif len(cond_stack) == 1 and cond_stack[0] == "#ifdef CLIPPY\n":
                    macros.load_preproc(fn, entry)
            continue
        if entry["type"].startswith("DEFPY") or (
            all_defun and entry["type"].startswith("DEFUN")
        ):
            if len(entry["args"][0]) != 1:
                sys.stderr.write(
                    "%s:%d: DEFPY function name not parseable (%r)\n"
                    % (fn, entry["lineno"], entry["args"][0])
                )
                errors += 1
                continue

            cmddef = entry["args"][2]
            cmddefx = []
            for i in cmddef:
                while i in macros:
                    i = macros[i]
                if i.startswith('"') and i.endswith('"'):
                    cmddefx.append(i[1:-1])
                    continue

                sys.stderr.write(
                    "%s:%d: DEFPY command string not parseable (%r)\n"
                    % (fn, entry["lineno"], cmddef)
                )
                errors += 1
                cmddefx = None
                break
            if cmddefx is None:
                continue
            cmddef = "".join([i for i in cmddefx])

            graph = clippy.Graph(cmddef)
            args = OrderedDict()
            always_args = set()
            for token, depth in clippy.graph_iterate(graph):
                if token.type not in handlers:
                    continue
                if token.varname is None:
                    continue
                arg = args.setdefault(token.varname, [])
                arg.append(handlers[token.type](token))
                always_args.add(token.varname)

            get_always_args(graph.first(), always_args)

            # print('-' * 76)
            # pprint(entry)
            # clippy.dump(graph)
            # pprint(args)

            params = {"cmddef": cmddef, "fnname": entry["args"][0][0]}
            argdefs = []
            argdecls = []
            arglist = []
            argblocks = []
            argassert = []
            doc = []
            canfail = 0

            def do_add(handler, basename, varname, attr=""):
                argdefs.append(",\\\n\t%s %s%s" % (handler.argtype, varname, attr))
                argdecls.append(
                    "\t%s\n"
                    % (
                        handler.decl.substitute({"varname": varname}).replace(
                            "\n", "\n\t"
                        )
                    )
                )
                arglist.append(", %s%s" % (handler.deref, varname))
                if basename in always_args and handler.canassert:
                    argassert.append(
                        """\tif (!%s) {
\t\tvty_out(vty, "Internal CLI error [%%s]\\n", "%s");
\t\treturn CMD_WARNING;
\t}\n"""
                        % (varname, varname)
                    )
                if attr == "":
                    at = handler.argtype
                    if not at.startswith("const "):
                        at = ". . . " + at
                    doc.append(
                        "\t%-26s %s  %s"
                        % (at, "alw" if basename in always_args else "opt", varname)
                    )

            for varname in args.keys():
                handler = mix_handlers(args[varname])
                # print(varname, handler)
                if handler is None:
                    continue
                do_add(handler, varname, varname)
                code = handler.code.substitute({"varname": varname}).replace(
                    "\n", "\n\t\t\t"
                )
                if handler.canfail:
                    canfail = 1
                strblock = ""
                if not handler.drop_str:
                    do_add(
                        StringHandler(None),
                        varname,
                        "%s_str" % (varname),
                        " __attribute__ ((unused))",
                    )
                    strblock = "\n\t\t\t%s_str = argv[_i]->arg;" % (varname)
                argblocks.append(
                    argblock.substitute(
                        {"varname": varname, "strblock": strblock, "code": code}
                    )
                )

            if dumpfd is not None:
                if len(arglist) > 0:
                    dumpfd.write('"%s":\n%s\n\n' % (cmddef, "\n".join(doc)))
                else:
                    dumpfd.write('"%s":\n\t---- no magic arguments ----\n\n' % (cmddef))

            params["cond_begin"] = "".join(cond_stack)
            params["cond_end"] = "".join(["#endif\n"] * len(cond_stack))
            params["argdefs"] = "".join(argdefs)
            params["argdecls"] = "".join(argdecls)
            params["arglist"] = "".join(arglist)
            params["argblocks"] = "".join(argblocks)
            params["canfail"] = canfail
            params["nonempty"] = len(argblocks)
            params["argassert"] = "".join(argassert)
            ofd.write(templ.substitute(params))

    return errors


if __name__ == "__main__":
    import argparse

    argp = argparse.ArgumentParser(description="FRR CLI preprocessor in Python")
    argp.add_argument(
        "--all-defun",
        action="store_const",
        const=True,
        help="process DEFUN() statements in addition to DEFPY()",
    )
    argp.add_argument(
        "--show",
        action="store_const",
        const=True,
        help="print out list of arguments and types for each definition",
    )
    argp.add_argument("-o", type=str, metavar="OUTFILE", help="output C file name")
    argp.add_argument("cfile", type=str)
    args = argp.parse_args()

    dumpfd = None
    if args.o is not None:
        ofd = StringIO()
        if args.show:
            dumpfd = sys.stdout
    else:
        ofd = sys.stdout
        if args.show:
            dumpfd = sys.stderr

    basepath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    macros = Macros()
    macros.load("lib/route_types.h")
    macros.load(os.path.join(basepath, "lib/command.h"))
    macros.load(os.path.join(basepath, "bgpd/bgp_vty.h"))
    # sigh :(
    macros.setup("PROTO_REDIST_STR", "FRR_REDIST_STR_ISISD")
    macros.setup("PROTO_IP_REDIST_STR", "FRR_IP_REDIST_STR_ISISD")
    macros.setup("PROTO_IP6_REDIST_STR", "FRR_IP6_REDIST_STR_ISISD")

    errors = process_file(args.cfile, ofd, dumpfd, args.all_defun, macros)
    if errors != 0:
        sys.exit(1)

    if args.o is not None:
        clippy.wrdiff(
            args.o, ofd, [args.cfile, os.path.realpath(__file__), sys.executable]
        )



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC
#
# January 23 2024, Christian Hopps <chopps@labn.net>
#
# Copyright (c) 2024, LabN Consulting, L.L.C.
#
"""
Test Traditional YANG Notifications
"""
import json
import os

import pytest
from lib.topogen import Topogen
from lib.topotest import json_cmp
from oper import check_kernel_32

pytestmark = [pytest.mark.ripd, pytest.mark.staticd, pytest.mark.mgmtd]

CWD = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture(scope="module")
def tgen(request):
    "Setup/Teardown the environment and provide tgen argument to tests"

    topodef = {
        "s1": ("r1", "r2"),
    }

    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for _, router in router_list.items():
        router.load_frr_config("frr.conf")

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


def test_frontend_notification(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"].net

    check_kernel_32(r1, "11.11.11.11", 1, "")

    fe_client_path = CWD + "/../lib/fe_client.py"
    rc, _, _ = r1.cmd_status(fe_client_path + " --help")

    if rc:
        pytest.skip("No protoc or present cannot run test")

    # Update config to non-matching authentication.
    conf = """
    conf t
    interface r1-eth0
    ip rip authentication string bar
    """
    r1.cmd_raises("vtysh", stdin=conf)

    try:
        output = r1.cmd_raises(
            fe_client_path + " --listen /frr-ripd:authentication-failure"
        )

        jsout = json.loads(output)
        expected = {"frr-ripd:authentication-failure": {"interface-name": "r1-eth0"}}
        result = json_cmp(jsout, expected)
        assert result is None

        output = r1.cmd_raises(
            fe_client_path + " --use-protobuf --listen /frr-ripd:authentication-failure"
        )
        jsout = json.loads(output)
        expected = {"frr-ripd:authentication-failure": {"interface-name": "r1-eth0"}}
        result = json_cmp(jsout, expected)
        assert result is None
    finally:
        # Update config to matching authentication.
        conf = """
        conf t
        interface r1-eth0
        ip rip authentication string foo
        """
        r1.cmd_raises("vtysh", stdin=conf)


def test_frontend_all_notification(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"].net

    check_kernel_32(r1, "11.11.11.11", 1, "")

    fe_client_path = CWD + "/../lib/fe_client.py"
    rc, _, _ = r1.cmd_status(fe_client_path + " --help")

    if rc:
        pytest.skip("No protoc or present cannot run test")

    # Update config to non-matching authentication.
    conf = """
    conf t
    interface r1-eth0
    ip rip authentication string bar
    """
    r1.cmd_raises("vtysh", stdin=conf)

    try:
        # The first notifications is a frr-ripd:authentication-type-failure
        # All the rest are frr-ripd:authentication-failure so we check for both.
        output = r1.cmd_raises(fe_client_path + " --listen /")
        jsout = json.loads(output)
        expected = {
            "frr-ripd:authentication-type-failure": {"interface-name": "r1-eth0"}
        }
        result = json_cmp(jsout, expected)
        if result is not None:
            expected = {
                "frr-ripd:authentication-failure": {"interface-name": "r1-eth0"}
            }
            result = json_cmp(jsout, expected)
        assert result is None

        output = r1.cmd_raises(fe_client_path + " --use-protobuf --listen /")
        jsout = json.loads(output)
        expected = {"frr-ripd:authentication-failure": {"interface-name": "r1-eth0"}}
        result = json_cmp(jsout, expected)
        assert result is None
    finally:
        # Update config to matching authentication.
        conf = """
        conf t
        interface r1-eth0
        ip rip authentication string foo
        """
        r1.cmd_raises("vtysh", stdin=conf)


def test_backend_yang_notification(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"].net

    check_kernel_32(r1, "11.11.11.11", 1, "")

    be_client_path = "/usr/lib/frr/mgmtd_testc"
    rc, _, _ = r1.cmd_status(be_client_path + " --help")

    if rc:
        pytest.skip("No mgmtd_testc")

    # Update config to non-matching authentication.
    conf = """
    conf t
    interface r1-eth0
    ip rip authentication string bar
    """
    r1.cmd_raises("vtysh", stdin=conf)

    try:
        output = r1.cmd_raises(
            be_client_path
            + " --timeout 20 --log file:mgmt_testc.log --listen /frr-ripd"
        )
        jsout = json.loads(output)
        expected = {"frr-ripd:authentication-failure": {"interface-name": "r1-eth0"}}
        result = json_cmp(jsout, expected)
        assert result is None
    finally:
        # Update config to matching authentication.
        conf = """
        conf t
        interface r1-eth0
        ip rip authentication string foo
        """
        r1.cmd_raises("vtysh", stdin=conf)



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bfd_profiles_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2020 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_bfd_profiles_topo1.py: Test the FRR BFD profile protocol integration.
"""

import os
import sys
import json
from functools import partial
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

pytestmark = [pytest.mark.bfdd, pytest.mark.bgpd, pytest.mark.isisd, pytest.mark.ospfd]


def setup_module(mod):
    "Sets up the pytest environment"

    topodef = {
        "s1": ("r1", "r2"),
        "s2": ("r2", "r3"),
        "s3": ("r3", "r4"),
        "s4": ("r4", "r5"),
        "s5": ("r1", "r6"),
    }
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for rname, router in router_list.items():
        daemon_file = "{}/{}/bfdd.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_BFD, daemon_file)

        daemon_file = "{}/{}/bgpd.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_BGP, daemon_file)

        daemon_file = "{}/{}/isisd.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_ISIS, daemon_file)

        daemon_file = "{}/{}/ospfd.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_OSPF, daemon_file)

        daemon_file = "{}/{}/ospf6d.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_OSPF6, daemon_file)

        daemon_file = "{}/{}/zebra.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_ZEBRA, daemon_file)

    # Initialize all routers.
    tgen.start_router()


def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def test_wait_protocols_convergence():
    "Wait for all protocols to converge"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for protocols to converge")

    def expect_loopback_route(router, iptype, route, proto):
        "Wait until route is present on RIB for protocol."
        logger.info("waiting route {} in {}".format(route, router))
        test_func = partial(
            topotest.router_json_cmp,
            tgen.gears[router],
            "show {} route json".format(iptype),
            {route: [{"protocol": proto}]},
        )
        _, result = topotest.run_and_expect(test_func, None, count=130, wait=1)
        assertmsg = '"{}" OSPF convergence failure'.format(router)
        assert result is None, assertmsg

    # Wait for R1 <-> R6 convergence.
    expect_loopback_route("r1", "ip", "10.254.254.6/32", "ospf")

    # Wait for R6 <-> R1 convergence.
    expect_loopback_route("r6", "ip", "10.254.254.1/32", "ospf")

    # Wait for R2 <-> R3 convergence.
    expect_loopback_route("r2", "ip", "10.254.254.3/32", "bgp")

    # Wait for R3 <-> R2 convergence.
    expect_loopback_route("r3", "ip", "10.254.254.2/32", "bgp")

    # Wait for R3 <-> R4 convergence.
    expect_loopback_route("r3", "ipv6", "2001:db8:3::/64", "isis")

    # Wait for R4 <-> R3 convergence.
    expect_loopback_route("r4", "ipv6", "2001:db8:1::/64", "isis")

    # Wait for R4 <-> R5 convergence.
    expect_loopback_route("r4", "ipv6", "2001:db8:3::/64", "ospf6")

    # Wait for R5 <-> R4 convergence.
    expect_loopback_route("r5", "ipv6", "2001:db8:2::/64", "ospf6")


def test_bfd_profile_values():
    "Assert that the BFD peers can find themselves."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for bfd peers to go up and checking profile values")

    for router in tgen.routers().values():
        json_file = "{}/{}/bfd-peers-initial.json".format(CWD, router.name)
        expected = json.loads(open(json_file).read())
        test_func = partial(
            topotest.router_json_cmp, router, "show bfd peers json", expected
        )
        _, result = topotest.run_and_expect(test_func, None, count=12, wait=0.5)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
